//////////////////////////////////////////////////////////////////////////
//
// Finnish KKJ Projection class & the Finnish national grid coordinate
// system.
//
// Version: 0.1
// Updated: 17.01.2002
// Updates: The code was preserved (17.1.2002 -Florian)
//
//////////////////////////////////////////////////////////////////////////

// OLD HEADER!
// Finnish KKJ Projection  class
//
// @author     Alexander Kolesnikov
// @created    18 marraskuuta 1999
// Revised    9 April, 2001
//
// The Finnish national grid coordinate system
//
// KKJ class extends Transverse Mercator class;
//
// Datum:         "European 1950"  or "WGS84"
// Scale factor:  1.0;
// Four Zones:    21, 24, 27 & 30 degrees;
// Pascal coding   Pavel Kopylov (18.04.2001)

Unit UnitKKJ;

Interface

Uses
  UnitMercator, Defines;

type
  TKKJ = class(TMercator)
  public
    Constructor Create;
    Destructor  Destroy; override;

    procedure ForwardTransform(Coordinate:TCoordinate; Var Projected:TProjected);
    procedure InverseTransform(Projected:TProjected; Var Coordinate:TCoordinate);
    function  CalcZoneNum(Coordinate:TCoordinate):Integer;
  private
    // Correction Type for WGS84  <--> KKJ
    // 0 -  No correction;             Datum = ED50
    // 1 -  Correction; Datum = WGS84
    // Default value: 1
    nCorrection             : Integer;
  End;


Implementation

//////////////////////////////////////////////////////////////////////////
// Constructor & Destructor
//////////////////////////////////////////////////////////////////////////
constructor TKKJ.Create;
begin
  inherited Create;
  nCorrection := 1;
end;


Destructor TKKJ.Destroy;
begin
  inherited Destroy;
end;


//////////////////////////////////////////////////////////////////////////
// Latitude, Longitude -> meters (Northing/Easting)
//////////////////////////////////////////////////////////////////////////
procedure TKKJ.ForwardTransform(Coordinate:TCoordinate; Var Projected:TProjected);
Var
  lat, lon                  : Double;
  scale_fact                : Double;
  northVal, eastVal         : Double;
  lon_origin                : Double;
  north_false, east_false   : Double;
  datum                     : TMapDatum;
  coord_1                   : TCoordinate;
  north, east, DX, DY, a, b : Double;
  proj_tmp                  : TProjected;
  intZoneNum                : Integer;
Begin
  lat := Coordinate.Latitude;
  lon := Coordinate.Longitude;
  scale_fact := 1;

  //
  north_false := 0;     // False Northing;

  //Make sure the longitude is between -180.00 .. 179.9
  lon  := (lon+180) - trunc((lon+180)/360)*360-180;

  datum := International;
  if (nCorrection = 0) Then datum := International;
  if (nCorrection = 1) Then datum := WGS84;

  intZoneNum := calcZoneNum(Coordinate);

  // Calc false easting;
  // 1 500 000; 2 500 000; 3 500 000; 4 500 000;
  east_false := 1500000 + (intZoneNum-1)*1000000;

  // Set origin of the Zone;
  lon_origin := 21 + (intZoneNum-1)*3;

  coord_1.Latitude  := lat;
  coord_1.Longitude := lon;
  coord_1.Datum     := datum;

  ///////////////////////////////////////////////////
  try
    forwardConversion(coord_1, lon_origin, proj_tmp);
    eastVal  := east_false  + scale_fact * proj_tmp.Easting;
    northVal := north_false + scale_fact * proj_tmp.Northing;
   except
     on EConversionInvalid do exit;
  End;

  north := 0; east := 0;

  //////////////////////////////////////////////
  // 0. No correction
  if (nCorrection = 0) Then
  Begin
    north := northVal;
    east  := eastVal;
  End;

  // 1. Correction;
  if (nCorrection = 1) Then
  Begin
    DX := 132.858;      DY := 132.490;
    a  := 1.0000021251; b  := 0.0000044173;
    north := DX + a*northVal - b*eastVal;
    east  := DY + b*northVal + a*eastVal;
  End;

  Projected.Northing   := north;
  Projected.Easting    := east;
  Projected.Datum      := International;
  Projected.ZoneNumber := intZoneNum;
End;



//////////////////////////////////////////////////////////////////////////
// meters (Northing/Easting) -> Latitude, Longitude
//////////////////////////////////////////////////////////////////////////
procedure TKKJ.InverseTransform(Projected:TProjected; Var Coordinate:TCoordinate);
Var
  north, east               : Double;
  scale_fact                : Double;
  lat_origin, lon_origin    : Double;
  north_false, east_false   : Double;
  datum                     : TMapDatum;
  A,B,n1,e1,a1,b1,c,dN,dE   : Double;
  proj_tmp                  : TProjected;
  intZoneNum                : Integer;        // Zone: 1..7;
Begin
  // Input parameters;
  north      := Projected.Northing;   // Northing;
  east       := Projected.Easting;    // Easting;
  intZoneNum := Projected.ZoneNumber;
  scale_fact := 1.0;                  // scale factor

  lat_origin := 0.0;
  north_false:= 0.0;

  // Set Zone parameters;
  lon_origin := 21 + 3*(abs(intZoneNum)-1);
  east_false := 1500000 + (intZoneNum-1)*1000000;

  // 2D Rotation
  datum := International;
  if (nCorrection = 0) Then datum := International;
  if (nCorrection = 1) Then
  Begin
    datum := WGS84;
    a1 := 1.0000021251; dN := 132.858;
    b1 := 0.0000044173; dE := 132.490;

    c  := a1*a1 + b1*b1;
    B  := b1/c;
    A  := a1/c;
    n1 :=  A*(north - dN) + B*(east - dE);
    e1 := -B*(north - dN) + A*(east - dE);

    north  := n1;
    east   := e1;
  End;

  north := north - north_false;
  east  := east - east_false;

  proj_tmp.Northing   := north;
  proj_tmp.Easting    := east;
  proj_tmp.Datum      := datum;
  Proj_tmp.ZoneNumber := 0;

  try
    inverseConversion(proj_tmp, lat_origin, lon_origin, scale_fact, Coordinate);
  except
     on EConversionInvalid do exit;
  End;
End;


//////////////////////////////////////////////////////////////////////////
// N Zone calculator - obsolete for YKJ
//////////////////////////////////////////////////////////////////////////
function TKKJ.calcZoneNum(Coordinate:TCoordinate):Integer;
Var
  zone                      : Integer;
Begin
  zone := trunc((Coordinate.Longitude - 19.5)/3) + 1;
  if (zone < 1) or (zone > 4) Then zone := -1;
  result := zone;
End;

End.
