//////////////////////////////////////////////////////////////////////////
//
// Transverse Mercator Projection
// The parent class for UTM, KKJ, YKJ
//
// Version: 0.1
// Updated: 17.01.2002
// Updates: The code was preserved (17.1.2002 -Florian)
//
//////////////////////////////////////////////////////////////////////////


// OLD HEADER!
// Transverse Mercator Projection
// The parent class for UTM, KKJ, YKJ
// Author          Alexander Kolesnikov 1999
// Created         8. marraskuuta 1999
// Pascal coding   Pavel Kopylov (18.04.2001)

Unit UnitMercator;

Interface

Uses
  Math, sysutils,
  Defines;

Type
  TMercator = class
  public
    Constructor Create;
    Destructor  Destroy; override;
    procedure   forwardConversion(Coordinate:TCoordinate; lon_center:Double;Var Projected:TProjected);
    procedure   inverseConversion(Projected:TProjected; lat_origin, lon_center, scale_fact: Double; Var Coordinate:TCoordinate);

  private
    MAX_VAL                 : Integer;
    MAXLONG                 : Integer;
    DBLLONG                 : Double;
    D2R                     : Double;
    R2D                     : Double;
    EPSLN                   : Double;

    function asinz(Con : Double):Double;
    function signum(arg:Double):Integer;
  End;

implementation

//////////////////////////////////////////////////////////////////////////
// Constructor & Destructor
//////////////////////////////////////////////////////////////////////////
constructor TMercator.Create;
begin
  MAX_VAL := 4;
  MAXLONG := 2147483647;
  DBLLONG := 4.61168601e18;
  D2R     := PI/180;
  R2D     := 180/PI;
  EPSLN   := 1.0e-10;        // precision of calc
end;

Destructor TMercator.Destroy;
begin
  MAX_VAL := 0;
  MAXLONG := 0;
  DBLLONG := 0;
  D2R     := 0;
  R2D     := 0;
  EPSLN   := 0;
end;


//////////////////////////////////////////////////////////////////////////
// Forvard Conversion
//////////////////////////////////////////////////////////////////////////
procedure TMercator.ForwardConversion(Coordinate:TCoordinate; lon_center:Double;Var Projected:TProjected);
Var
  East, North               : Double;
  r_major, r_minor          : Double;
  lat, lon                  : Double;
  temp, es, esp             : Double;
  N, T, C, A                : Double;
  A2, A3, es2, es3          : Double;
  M                         : Double;
Begin
  // Get Ellipsoid parameters;
  r_major := Ellipsoid[Coordinate.Datum].Major;
  r_minor := Ellipsoid[Coordinate.Datum].Minor;
  temp    := r_minor/r_major;
  es      := 1 - temp*temp;
  esp     := es/(1.0 - es);
  lat     := Coordinate.Latitude;
  lon     := Coordinate.Longitude;

  //Make sure the longitude is between -180.00 .. 179.9
  lon := (lon+180)- trunc((lon+180)/360)*360-180;
  lat := lat * D2R;
  lon := lon * D2R;
  lon_center := lon_center * D2R;

  N  := r_major / sqrt(1-es*sin(lat)*sin(lat));
  T  := tan(lat)*tan(lat);
  C  := esp*cos(lat)*cos(lat);
  A  := cos(lat)*(lon - lon_center);

  A2 := A*A;
  A3 := A*A2;
  es2:= es*es;
  es3:= es*es2;

  M  := r_major*( (1 - es/4 - 3*es2/64 - 5*es3/256)*lat
                 -(3*es/8 + 3*es2/32 + 45*es3/1024)*sin(2*lat)
                 +(15*es2/256 + 45*es3/1024)*sin(4*lat)
                 -(35*es3/3072)*sin(6*lat));

  east  := N*(A+(1-T+C)*A3/6 + (5.0-18.0*T+T*T+72*C-58*esp)*A2*A3/120);
  north := M + N*tan(lat)*(A2/2+(5-T+9*C+4*C*C)*A2*A2/24.0
             + (61.0-58.0*T+T*T+600.0*C-330.0*esp)*A3*A3/720.0);

  Projected.Northing   := north;
  Projected.Easting    := east;
  Projected.Datum      := Coordinate.Datum;
End;


//////////////////////////////////////////////////////////////////////////
// Inverse Conversion
//////////////////////////////////////////////////////////////////////////
procedure TMercator.InverseConversion(Projected:TProjected; lat_origin, lon_center, scale_fact: Double; Var Coordinate:TCoordinate);
Var
  lat, lon                  : Double;
  dEast, dNorth             : Double;
  max_iter                  : LongInt;
  r_major, r_minor          : Double;
  con,phi,delta_phi         : Double;
  sin_phi, cos_phi, tan_phi, c, cs, t, ts, n, r, d, ds, lon_tt   : Double;
  ff,h,g, e, es, e0,e1,e2,e3,ml0,esp,temp, m10, cc,dd,qq       : Double;
  ind, i                    : Integer;
  Failure                   : Boolean;
Begin
  max_iter := 8;             // maximun number of iterations
  m10 := 0;

  dEast  := Projected.Easting;
  dNorth := Projected.Northing;

  r_major := Ellipsoid[Projected.Datum].Major;
  r_minor := Ellipsoid[Projected.Datum].Minor;

  // Degrees to radians
  lat_origin := lat_origin*D2R;
  lon_center := lon_center*D2R;

  temp := r_minor / r_major;
  es := 1.0 - temp*temp;
  e  := sqrt(es);

  e0 := 1.0 - 0.25*es*(1.0 + es/16.0*(3.0 + 1.25*es));
  e1 := 0.375*es*(1.0 + 0.25*es*(1.0 + 0.46875*es));
  e2 := 0.05859375*es*es*(1.0 + 0.75*es);
  e3 := es*es*es*(35.0/3072.0);

  dd := e0*lat_origin-e1*sin(2*lat_origin)+e2*sin(4*lat_origin) - e3*sin(6*lat_origin);

  ml0 := r_major * dd;

  esp := es / (1.0 - es);

  if (es < 0.00001) Then ind := 1  // Spherical
  else                   ind := 0; // Elliptical

  if(ind <> 0) Then
  Begin                            // Spherical form;
    ff   := exp( dEast/(r_major * scale_fact) );
    g    := 0.5 * (ff - 1/ff);
    temp := lat_origin + dNorth/(r_major * scale_fact);
    h    := cos(temp);
    con  := sqrt((1.0 - h*h)/(1.0 + g*g));
    lat  := asinz(con);
    if(temp < 0) Then lat := -lat;

    if (g = 0) and (h = 0) Then lon := lon_center
    Else
    Begin
      lon_tt := ArcTan2(g,h) + lon_center;
      lon    := (lon_tt+180) - trunc((lon_tt+180)/360)*360-180;
    End;
  End
  Else                             // Ellipsoid form;
  Begin
    con := (m10 + dNorth/scale_fact) / r_major;
    phi := con;

    i := 0;
    Failure := False;
    repeat
      delta_phi := ((con + e1*sin(2.0*phi)-e2*sin(4.0*phi)+e3*sin(6.0*phi)) / e0) - phi;
      phi := phi + delta_phi;

      if (i >= max_iter) Then Failure := True;
      inc(i);
    Until (abs(delta_phi) <= EPSLN) or (Failure);

    if Failure Then
    Begin
      raise EConversionInvalid.Create('Latitude is failed to converge');
      exit;
    End;

    if(abs(phi) < PI/2) Then
    Begin
      sin_phi := sin(phi);
      cos_phi := cos(phi);
      tan_phi := tan(phi);

      c   := esp * cos_phi*cos_phi;
      cs  := c*c;
      t   := tan_phi*tan_phi;
      ts  := t*t;
      con := 1.0 - es * sin_phi*sin_phi;
      n   := r_major / sqrt(con);
      r   := n * (1.0 - es) / con;
      d   := dEast / (n * scale_fact);
      ds  := d*d;

      lat := phi - (n*tan_phi*ds/r)*(0.5-ds/24.0* (5.0+3.0*t+10.0*c-4.0*cs -
                    9.0*esp-ds/30.0*(61.0+90.0*t+298.0*c+45.0*ts-252.0*esp-3.0*cs)));

      lon_tt := lon_center+(d*(1.0-ds/6.0*(1.0+2.0*t+
                 c-ds/20.0*(5.0-2.0*c+28.0*t-3.0*cs+8.0*esp+24.0*ts)))/cos_phi);

      lon    := (lon_tt+180) - trunc((lon_tt+180)/360)*360-180;
    End
    Else
    Begin
      lat := PI/2 * signum(dNorth);
      lon := lon_center;
    End
  End;                             // ind==0

  lat := lat * R2D;
  lon := lon * R2D;

  Coordinate.Latitude  := lat;
  Coordinate.Longitude := lon;
  Coordinate.Datum     := Projected.Datum;
End;

//////////////////////////////////////////////////////////////////////////
//  eliminate roundoff errors in asin
//////////////////////////////////////////////////////////////////////////
function TMercator.asinz(Con:Double):Double;
Begin
  if (abs(con) > 1.0) Then
  Begin
    if (con > 1.0) Then con :=  1.0
    else                con := -1.0;
  End;
  Result := ArcSin(con);
End;


function TMercator.signum(arg:Double):Integer;
Begin
  Result := 0;
  if arg>0 Then Result := 1;
  if arg<0 Then Result := -1;
End;


End.
