//////////////////////////////////////////////////////////////////////////
//
// YKJ       (Finnish) Projection  class
// extends   TransverseMercator class
//
// Version: 0.1
// Updated: 17.01.2002
// Updates: The code was preserved (17.1.2002 -Florian)
//
//////////////////////////////////////////////////////////////////////////

// OLD HEADER!
// YKJ (Finnish) Projection  class
//
// @author     Alexander Kolesnikov
// @created    18 marraskuuta 1999
//
// Revised:
//   March 23, 2001;   Alexander Kolesnikov
//   April  9, 2001;   Alexander Kolesnikov
// YKJ class extends Transverse Mercator class;
// KartastoKoordinaattiJa"rjestelma" Projection;
//
//
// Datum  :   European datum 1950 (ED50) or  WGS84
// Scale factor:  1.0;
// One Zone    :  27 degrees;
// Yhtena"iskoordinaatisto
// Pascal coding   Pavel Kopylov (18.04.2001)

Unit UnitYKJ;

Interface

Uses
  UnitMercator, Defines;

type
  TYKJ = class(TMercator)
  public
    Constructor Create;
    Destructor  Destroy; override;

    procedure ForwardTransform(Coordinate:TCoordinate; Var Projected:TProjected);
    procedure InverseTransform(Projected:TProjected; Var Coordinate:TCoordinate);
    function  CalcZoneNum(Coordinate:TCoordinate):Integer;
  private

    // Correction Type for WGS84 <--> YKJ
    //   0 -  No correction;        Datum = ED50
    //   1 -  One step correction;  Datum = WGS84
    //   Default value: 1
    nCorrection                : Integer;
  End;


implementation

//////////////////////////////////////////////////////////////////////////
// Constructor & Destructor
//////////////////////////////////////////////////////////////////////////
constructor TYKJ.Create;
begin
  inherited Create;
  nCorrection := 1;
end;


Destructor TYKJ.Destroy;
begin
  nCorrection := 0;
  inherited Destroy;
end;


//////////////////////////////////////////////////////////////////////////
// Latitude, Longitude -> meters (Northing/Easting)
//////////////////////////////////////////////////////////////////////////
procedure TYKJ.ForwardTransform(Coordinate:TCoordinate; Var Projected:TProjected);
Var
  lat, lon                  : Double;
  scale_fact                : Double;
  northVal, eastVal         : Double;
  lon_origin                : Double;
  north_false, east_false   : Double;
  datum                     : TMapDatum;
  coord_1                   : TCoordinate;
  north, east, DX, DY, a, b : Double;
  proj_tmp                  : TProjected;

Begin
  lat := Coordinate.Latitude;
  lon := Coordinate.Longitude;
  scale_fact := 1;

  lon_origin  := 27.0;
  north_false := 0;         // False Northing;
  east_false  := 3500000;   // False Easting;

  //Make sure the longitude is between -180.00 .. 179.9
  lon := (lon+180) - trunc((lon+180)/360)*360-180;

  datum := International;
  if (nCorrection = 0) Then datum := International; // KKJ(YKJ) uses Hayford 1910 ellipsoid
  if (nCorrection = 1) Then datum := WGS84;

  coord_1.Latitude  := lat;
  coord_1.Longitude := lon;
  coord_1.Datum     := datum;

  ///////////////////////////////////////////////////
  try
    forwardConversion(coord_1, lon_origin, proj_tmp);
    eastVal  := east_false  + scale_fact * proj_tmp.Easting;
    northVal := north_false + scale_fact * proj_tmp.Northing;
   except
     on EConversionInvalid do exit;
  End;

  north:=0; east:=0;

  //////////////////////////////////////////////
  // 0. No correction
  if(nCorrection = 0) Then
  Begin
    north := northVal;
    east  := eastVal;
  End;

  // 4.1. 2D Rotation;
  if(nCorrection = 1) Then
  Begin
    DX := 132.858;      DY := 132.490;
    a  := 1.0000021251; b  := 0.0000044173;

    north := DX + a*northVal - b*eastVal;
    east  := DY + b*northVal + a*eastVal;
  End;

  Projected.Northing   := north;
  Projected.Easting    := east;
  Projected.Datum      := International;
  Projected.ZoneNumber := 0;
End;


//////////////////////////////////////////////////////////////////////////
// meters (Northing/Easting) -> Latitude, Longitude
//////////////////////////////////////////////////////////////////////////
procedure TYKJ.InverseTransform(Projected:TProjected; Var Coordinate:TCoordinate);
Var
  north, east               : Double;
  scale_fact                : Double;
  lat_origin, lon_origin    : Double;
  north_false, east_false   : Double;
  datum                     : TMapDatum;
  A,B,n1,e1,a1,b1,c,dN,dE   : Double;
  proj_tmp                  : TProjected;
Begin
  // Input parameters;
  north := Projected.Northing;   // Northing;
  east  := Projected.Easting;    // Easting;
  scale_fact := 1.0;             // scale factor

  lat_origin := 0.0;
  lon_origin := 27.0;
  east_false := 3500000.0;
  north_false:= 0.0;

  // 2D Rotation
  datum := International;
  if(nCorrection = 0) Then datum := International;
  if(nCorrection = 1) Then
  Begin
    datum := WGS84;
    a1 := 1.0000021251; dN := 132.858;
    b1 := 0.0000044173; dE := 132.490;

    c  := a1*a1 + b1*b1;
    B  := b1/c;
    A  := a1/c;
    n1 :=  A*(north - dN) + B*(east - dE);
    e1 := -B*(north - dN) + A*(east - dE);
    north := n1;
    east  := e1;
  End;

  north := north - north_false;
  east  := east - east_false;

  proj_tmp.Northing   := north;
  proj_tmp.Easting    := east;
  proj_tmp.Datum      := datum;
  Proj_tmp.ZoneNumber := 0;

  try
    InverseConversion(proj_tmp, lat_origin, lon_origin, scale_fact, Coordinate);
  except
     on EConversionInvalid do exit;
  End;
End;


//////////////////////////////////////////////////////////////////////////
// N Zone calculator - obsolete for YKJ
//////////////////////////////////////////////////////////////////////////
function TYKJ.calcZoneNum(Coordinate:TCoordinate):Integer;
Var
  zone                      : Integer;
Begin
  zone := trunc((Coordinate.Longitude - 19.5)/3) + 1;
  if (zone < 1) or (zone > 4) Then zone := -1;
  result := zone;
End;

End.
